@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright 2010-2015 The pygit2 contributors
+# Copyright 2010-2017 The pygit2 contributors
 #
 # This file is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License, version 2,
@@ -32,6 +32,7 @@ from __future__ import absolute_import
 from string import hexdigits
 import sys, tarfile
 from time import time
+
 if sys.version_info[0] < 3:
     from cStringIO import StringIO
 else:
@@ -40,11 +41,12 @@ else:
 import six
 
 # Import from pygit2
-from _pygit2 import Repository as _Repository
+from _pygit2 import Repository as _Repository, init_file_backend
 from _pygit2 import Oid, GIT_OID_HEXSZ, GIT_OID_MINPREFIXLEN
 from _pygit2 import GIT_CHECKOUT_SAFE, GIT_CHECKOUT_RECREATE_MISSING, GIT_DIFF_NORMAL
 from _pygit2 import GIT_FILEMODE_LINK
-from _pygit2 import Reference, Tree, Commit, Blob
+from _pygit2 import GIT_BRANCH_LOCAL, GIT_BRANCH_REMOTE, GIT_BRANCH_ALL
+from _pygit2 import Reference, Tree, Commit, Blob, Signature
 
 from .config import Config
 from .errors import check_error
@@ -56,23 +58,11 @@ from .utils import to_bytes, is_string
 from .submodule import Submodule
 
 
-class Repository(_Repository):
-
-    def __init__(self, path, *args, **kwargs):
-        if not isinstance(path, six.string_types):
-            path = path.decode('utf-8')
-        super(Repository, self).__init__(path, *args, **kwargs)
+class BaseRepository(_Repository):
+    def __init__(self, backend, *args, **kwargs):
+        super(BaseRepository, self).__init__(backend, *args, **kwargs)
         self._common_init()
 
-    @classmethod
-    def _from_c(cls, ptr, owned):
-        cptr = ffi.new('git_repository **')
-        cptr[0] = ptr
-        repo = cls.__new__(cls)
-        super(cls, repo)._from_c(bytes(ffi.buffer(cptr)[:]), owned)
-        repo._common_init()
-        return repo
-
     def _common_init(self):
         self.remotes = RemoteCollection(self)
 
@@ -279,11 +269,6 @@ class Repository(_Repository):
         oid = reference.resolve().target
         treeish = self[oid]
         self.checkout_tree(treeish, **kwargs)
-        head = self.lookup_reference('HEAD')
-        if head.type == C.GIT_REF_SYMBOLIC:
-            from_ = self.head.shorthand
-        else:
-            from_ = head.target.hex
 
         self.set_head(refname)
 
@@ -323,19 +308,19 @@ class Repository(_Repository):
         Keyword arguments:
 
         a
-            None, a str (that refers to an Object, see revparse_single()) or a 
+            None, a str (that refers to an Object, see revparse_single()) or a
             Reference object.
             If None, b must be None, too. In this case the working directory is
             compared with the index. Otherwise the referred object is compared to
             'b'.
-        
+
         b
             None, a str (that refers to an Object, see revparse_single()) or a
             Reference object.
             If None, the working directory is compared to 'a'. (except
             'cached' is True, in which case the index is compared to 'a').
             Otherwise the referred object is compared to 'a'
-    
+
         cached
             if 'b' is None, by default the working directory is compared to 'a'.
             If 'cached' is set to True, the index/staging area is used for comparing.
@@ -500,6 +485,7 @@ class Repository(_Repository):
     @staticmethod
     def _merge_options(favor):
         """Return a 'git_merge_opts *'"""
+
         def favor_to_enum(favor):
             if favor == 'normal':
                 return C.GIT_MERGE_FILE_FAVOR_NORMAL
@@ -546,13 +532,13 @@ class Repository(_Repository):
             theirs._to_c() if theirs is not None else (ffi.NULL, ffi.NULL))
 
         err = C.git_merge_file_from_index(
-                cmergeresult, self._repo,
-                cancestor, cours, ctheirs,
-                ffi.NULL);
+            cmergeresult, self._repo,
+            cancestor, cours, ctheirs,
+            ffi.NULL);
         check_error(err)
 
         ret = ffi.string(cmergeresult.ptr,
-                cmergeresult.len).decode('utf-8')
+                         cmergeresult.len).decode('utf-8')
         C.git_merge_file_result_free(cmergeresult)
 
         return ret
@@ -674,7 +660,16 @@ class Repository(_Repository):
             consider. Increasing above 10 will take slightly longer but may
             produce a more accurate result. A value of 0 will cause only exact
             matches to be output.
-        :param int describe_strategy: A GIT_DESCRIBE_* constant.
+        :param int describe_strategy: A `GIT_DESCRIBE_*` constant. Can be one
+            of:
+
+            * `GIT_DESCRIBE_DEFAULT` - Only match annotated tags. (This is
+              equivalent to setting this parameter to `None`.)
+            * `GIT_DESCRIBE_TAGS` - Match everything under refs/tags/
+              (includes lightweight tags).
+            * `GIT_DESCRIBE_ALL` - Match everything under refs/ (includes
+              branches).
+
         :param str pattern: Only consider tags matching the given `glob(7)`
             pattern, excluding the "refs/tags/" prefix.
         :param bool only_follow_first_parent: Follow only the first parent
@@ -706,7 +701,11 @@ class Repository(_Repository):
         if describe_strategy is not None:
             options.describe_strategy = describe_strategy
         if pattern:
-            options.pattern = ffi.new('char[]', to_bytes(pattern))
+            # The returned pointer object has ownership on the allocated
+            # memory. Make sure it is kept alive until git_describe_commit() or
+            # git_describe_workdir() are called below.
+            pattern_char = ffi.new('char[]', to_bytes(pattern))
+            options.pattern = pattern_char
         if only_follow_first_parent is not None:
             options.only_follow_first_parent = only_follow_first_parent
         if show_commit_oid_as_fallback is not None:
@@ -752,6 +751,96 @@ class Repository(_Repository):
         finally:
             C.git_describe_result_free(result[0])
 
+    #
+    # Stash
+    #
+    def stash(self, stasher, message=None, keep_index=False,
+              include_untracked=False, include_ignored=False):
+        """Save changes to the working directory to the stash.
+
+        :param Signature stasher: The identity of the person doing the stashing.
+        :param str message: An optional description of stashed state.
+        :param bool keep_index: Leave changes already added to the index
+            in the working directory.
+        :param bool include_untracked: Also stash untracked files.
+        :param bool include_ignored: Also stash ignored files.
+
+        :returns: The Oid of the stash merge commit.
+        :rtype: Oid
+
+        Example::
+
+            >>> repo = pygit2.Repsitory('.')
+            >>> repo.stash(repo.default_signature(), 'WIP: stashing')
+        """
+
+        if message is not None:
+            stash_msg = ffi.new('char[]', to_bytes(message)) if message else ffi.NULL
+        else:
+            stash_msg = ffi.NULL
+
+        flags = 0
+        flags |= keep_index * C.GIT_STASH_KEEP_INDEX
+        flags |= include_untracked * C.GIT_STASH_INCLUDE_UNTRACKED
+        flags |= include_ignored * C.GIT_STASH_INCLUDE_IGNORED
+
+        stasher_cptr = ffi.new('git_signature **')
+        ffi.buffer(stasher_cptr)[:] = stasher._pointer[:]
+
+        coid = ffi.new('git_oid *')
+        err = C.git_stash_save(coid, self._repo, stasher_cptr[0], stash_msg, flags)
+        check_error(err)
+
+        return Oid(raw=bytes(ffi.buffer(coid)[:]))
+
+    @staticmethod
+    def _stash_args_to_options(reinstate_index=False, **kwargs):
+        stash_opts = ffi.new('git_stash_apply_options *')
+        check_error(C.git_stash_apply_init_options(stash_opts, 1))
+
+        flags = reinstate_index * C.GIT_STASH_APPLY_REINSTATE_INDEX
+        stash_opts.flags = flags
+
+        copts, refs = Repository._checkout_args_to_options(**kwargs)
+        stash_opts.checkout_options = copts[0]
+
+        return stash_opts
+
+    def stash_apply(self, index=0, **kwargs):
+        """Apply a stashed state in the stash list to the working directory.
+
+        :param int index: The position within the stash list of the stash to apply.
+            0 is the most recent stash.
+        :param bool reinstate_index: Try to reinstate stashed changes to the index.
+
+        The checkout options may be customized using the same arguments taken by
+        Repository.checkout().
+
+        Example::
+
+            >>> repo = pygit2.Repsitory('.')
+            >>> repo.stash(repo.default_signature(), 'WIP: stashing')
+            >>> repo.stash_apply(strategy=GIT_CHECKOUT_ALLOW_CONFLICTS)
+        """
+        stash_opts = Repository._stash_args_to_options(**kwargs)
+        check_error(C.git_stash_apply(self._repo, index, stash_opts))
+
+    def stash_drop(self, index=0):
+        """Remove a stashed state from the stash list.
+
+        :param int index: The position within the stash list of the stash to remove.
+            0 is the most recent stash.
+        """
+        check_error(C.git_stash_drop(self._repo, index))
+
+    def stash_pop(self, index=0, **kwargs):
+        """Apply a stashed state and remove it from the stash list.
+
+        For arguments, see Repository.stash_apply().
+        """
+        stash_opts = Repository._stash_args_to_options(**kwargs)
+        check_error(C.git_stash_pop(self._repo, index, stash_opts))
+
     #
     # Utility for writing a tree into an archive
     #
@@ -812,11 +901,11 @@ class Repository(_Repository):
             info = tarfile.TarInfo(prefix + entry.path)
             info.size = len(content)
             info.mtime = timestamp
-            info.uname = info.gname = 'root' # just because git does this
+            info.uname = info.gname = 'root'  # just because git does this
             if entry.mode == GIT_FILEMODE_LINK:
                 info.type = tarfile.SYMTYPE
                 info.linkname = content.decode("utf-8")
-                info.mode = 0o777 # symlinks get placeholder
+                info.mode = 0o777  # symlinks get placeholder
                 info.size = 0
                 archive.addfile(info)
             else:
@@ -920,3 +1009,130 @@ class Repository(_Repository):
 
         err = C.git_repository_set_ident(self._repo, to_bytes(name), to_bytes(email))
         check_error(err)
+
+    def revert_commit(self, revert_commit, our_commit, mainline=0):
+        """Reverts the given Commit against the given "our" Commit,
+           producing an Index that reflects the result of the revert.
+
+        Arguments
+
+        revert_commit
+            The Commit to revert
+        our_commit
+            The Commit to revert against (eg, HEAD)
+        mainline
+            The parent of the revert Commit, if it is a merge (i.e. 1, 2)
+
+        Returns an Index with the result of the revert.
+        """
+        cindex = ffi.new('git_index **')
+        revert_commit_ptr = ffi.new('git_commit **')
+        our_commit_ptr = ffi.new('git_commit **')
+
+        ffi.buffer(revert_commit_ptr)[:] = revert_commit._pointer[:]
+        ffi.buffer(our_commit_ptr)[:] = our_commit._pointer[:]
+
+        opts = ffi.new('git_merge_options *')
+        err = C.git_merge_init_options(opts, C.GIT_MERGE_OPTIONS_VERSION)
+        check_error(err)
+
+        err = C.git_revert_commit(
+            cindex, self._repo, revert_commit_ptr[0], our_commit_ptr[0], mainline, opts
+        )
+        check_error(err)
+
+        return Index.from_c(self, cindex)
+
+
+class Branches(object):
+    def __init__(self, repository, flag=GIT_BRANCH_ALL):
+        self._repository = repository
+        self._flag = flag
+
+        if flag == GIT_BRANCH_ALL:
+            self.local = Branches(repository, flag=GIT_BRANCH_LOCAL)
+            self.remote = Branches(repository, flag=GIT_BRANCH_REMOTE)
+
+    def __getitem__(self, name):
+        branch = None
+        if self._flag & GIT_BRANCH_LOCAL:
+            branch = self._repository.lookup_branch(name, GIT_BRANCH_LOCAL)
+
+        if branch is None and self._flag & GIT_BRANCH_REMOTE:
+            branch = self._repository.lookup_branch(name, GIT_BRANCH_REMOTE)
+
+        if branch is None:
+            raise KeyError('Branch not found: {}'.format(name))
+
+        return branch
+
+    def get(self, key):
+        try:
+            return self[key]
+        except KeyError:
+            return None
+
+    def __iter__(self):
+        for branch_name in self._repository.listall_branches(self._flag):
+            yield branch_name
+
+    def create(self, name, commit, force=False):
+        return self._repository.create_branch(name, commit, force)
+
+    def delete(self, name):
+        self[name].delete()
+
+    def __contains__(self, name):
+        return self.get(name) is not None
+
+
+class References(object):
+    def __init__(self, repository):
+        self._repository = repository
+
+    def __getitem__(self, name):
+        return self._repository.lookup_reference(name)
+
+    def get(self, key):
+        try:
+            return self[key]
+        except KeyError:
+            return None
+
+    def __iter__(self):
+        for ref_name in self._repository.listall_references():
+            yield ref_name
+
+    def create(self, name, target, force=False):
+        return self._repository.create_reference(name, target, force)
+
+    def delete(self, name):
+        self[name].delete()
+
+    def __contains__(self, name):
+        return self.get(name) is not None
+
+    @property
+    def objects(self):
+        return self._repository.listall_reference_objects()
+
+
+class Repository(BaseRepository):
+    def __init__(self, path, *args, **kwargs):
+        if not isinstance(path, six.string_types):
+            path = path.decode('utf-8')
+
+        path_backend = init_file_backend(path)
+        super(Repository, self).__init__(backend=path_backend, *args, **kwargs)
+
+        self.branches = Branches(self)
+        self.references = References(self)
+
+    @classmethod
+    def _from_c(cls, ptr, owned):
+        cptr = ffi.new('git_repository **')
+        cptr[0] = ptr
+        repo = cls.__new__(cls)
+        super(cls, repo)._from_c(bytes(ffi.buffer(cptr)[:]), owned)
+        repo._common_init()
+        return repo
